/*************************     luc2nts.c    *********************************

	Converts a merged Neurolucida file to Eutectic NTS format for
	use with ntscable software. Takes a source file name as an argument.
	Creates a target file with same root name as source but with an .nts
	extension. Neurolucida header data is conserved.

	Ignores NTS major codes 0, 3, 5, 33 or 34.
	Ignores NTS minor codes 6, 9, 10, 11, 13, 14, 30, 31, >108

	Written by Gerald Oakson and Alain Destexhe, Laval University, 1996

*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

struct {
	char    code_in[8];     /* receives entire Neurolucida code string */
	int     majcode;        /* Neurolucida major code */
	int     mincode;        /* Neurolucida minor code */
	float   xval;           /* x coordinate */
	float   yval;           /* y coordinate */
	float   zval;           /* x coordinate */
	float   diam;           /* diameter */
	char    ntscode[3];     /* equivalent NTS code */
	char    fname_in[12];   /* source file name */
	char    fname_out[12];  /* target file name */
	char    data_type[14];  /* data type ID for MTO & soma codes */
} vars;

/************** globals ***************/
static  int     somaflag = 0, lineincr = 0;
static  long    pointnumber = 0;
static  char    result[68];
char nts_hdr[] =
"\n Point  Type  Tag      X       Y       Z  Thick   Name    Attachment\n\n";

/* 
" Point  Type  Tag      X       Y       Z  Thick   data_type    Attachment"; 
*/

/******* prototypes *********************/
void decodeline();
void encodeline();
void majcode1();
void majcode2();
void majcode10();
void majcode32();
void majcode33();


/********************  Main  ********************/
void main(argc, argv)
int     argc;
char    *argv[];
{       
	int     err, i = 0, ind, firstdataflag = 0, ferr;
	long    nem, linenumber = 0;
	FILE    *fp, *tp;       
	char    line[80], *lineptr, ntsheader[500], vers[11], *ch, *err1;
	float   *coordptr[4];
	size_t  mem, linelen = 64;
	
	coordptr[0] =&vars.xval; /* set coordinates pointer array */
	coordptr[1] =&vars.yval;
	coordptr[2] =&vars.zval;
	coordptr[3] =&vars.diam;

	if(argv[1] == NULL) {
		printf("No source file specified\nUsage:  luc2nts source_file\n");
		exit(-1);
	}
	strcpy(vars.fname_in, argv[1]);
	fp = fopen(vars.fname_in, "r");  /* open file */
	if(fp == NULL) {           
		printf("Cannot open source file %s\n", vars.fname_in);
		fclose(fp);
		exit(-1);        
	}

	/* create a target filename (change extension to .nts) */
	strcpy(vars.fname_out,vars.fname_in);
	ch = strchr(vars.fname_out,(int) '.');
	strcpy(ch + 1, "nts");

	nem = remove(vars.fname_out);       /* remove old target file if it exists */
	tp = fopen(vars.fname_out, "a+");   /* create new target file */
	printf("\nConverting NeuroLucida file: %s\n                to NTS file: %s \n",
					vars.fname_in, vars.fname_out);

	do {
		lineptr = line;
		ch = fgets(line, linelen, fp);
		linenumber ++;
		if(feof(fp) || (ferr = ferror(fp))) {  /* test result */
			if(0 != ferr) {
				printf(" error reading line %ld ", linenumber);
				break;
			}
			break;
		}

		if(linenumber == 1) {
			if(0 == strncmp(lineptr,"version ", 8)) {
				sprintf(ntsheader,"Neurolucida file %s to NTS format\n", vars.fname_in);
				continue;
			}
			else {
				printf("This is NOT a Neurolucida file");
				break;
			}
		}

		if(0 == strncmp(line,"\r", 1))  /* remove 1st char '\r' if present */
			lineptr ++;

		if(0 == strncmp(lineptr,"[", 1)) { /* is this a data line? */
			if(0 == firstdataflag) {
				strcat(ntsheader,nts_hdr);  /* add Eutectic header */
				nem = fputs(ntsheader, tp); /* write header before 1st data */
				firstdataflag = 1;
			}
			decodeline(lineptr, coordptr);    /* put data in struct vars */
			if(0 == lineincr) {         /* if no line skipping necessary */
				pointnumber ++;
				encodeline(coordptr, pointnumber); /* create new target line */
				if(0 != strcmp(vars.ntscode, "XXX")) { /* if code OK append to file */
					err = fputs(result, tp);
					if(err == EOF) {          /* check for write error */
						printf("error writing %s ", result);
						break;
					}
				}
					else
						pointnumber --;
				}
				else
					lineincr --;
			continue;
			}
		strcat(ntsheader,lineptr); /* if a hdr line or \n then append */
	}
	while(1);

	fclose(fp);     /* close files */
	fclose(tp);
}

/****************** decodeline *************************/
/* Places 1 data line's parameters in struct vars  */

void decodeline(line, coordptr)
char        *line;
float       *coordptr[4];
{                    
	int     err, i=0, ind, minor;
	char    delim = ',';    
	char    *ch, *lp, codestr[8];

	strcpy(vars.code_in, "        ");
	lp = line;
	ch = strchr(lp,(int) ']');   /* store code string */
	ind = 1 + ch - lp;      
	strncpy(vars.code_in, lp,(size_t) ind);

	ch = strchr(line,(int) ',');    /* get major code */
	lp ++;  ind = ch - lp;
	strcpy(codestr,"  ");
	strncpy(codestr, lp,(size_t) ind);
	vars.majcode = atoi(codestr);

	ch = strchr(line,(int) ']'); /* get minor code */    
	lp += ind + 1;  
	ind = ch - lp;  
	strcpy(codestr,"       ");
	strncpy(codestr, lp,(size_t) ind);
	vars.mincode = atoi(codestr);
	lp = 1 + strchr(line, (int) '(');

	for(i = 0; i < 4; i++)  {   /* get values x, y, z, & diam */
		if(i == 2)              /* default delimiter ',' for x & y */
			delim = ')';        /* delimiter for z value */
		if(i == 3)                      
			delim = '\n';       /* delimiter for diam */
		ch = strchr(lp,(int) delim);            
		ind = ch - lp;          
		strncpy(codestr, lp,(size_t) ind);
		*coordptr[i] = (float) atof(codestr);
		lp += ind + 1;
	}       
}

/******************* encodeline ************************/
/* Creates NTS formatted line from values in struct vars */

void encodeline(coordptr, pointnum)
float   *coordptr[4];
long     pointnum;
{      
	char    str[7];
	int     i;

	lineincr = 0;
	strcpy(vars.ntscode, "   ");
	strcpy(vars.data_type,"            ");
	sprintf(result,"%6ld  ",pointnum);  /* point number */
	switch(vars.majcode) {  /* select Neurolucida major code */
		case 1: {                       
			majcode1(vars.mincode);
			break;          
		}                                       
		case 2: {                       
			majcode2(vars.mincode);
			break;          
		}                                       
		case 10: {                      
			majcode10(vars.mincode);
			break;          
		}                                              
		default: {                      
			strcpy(vars.ntscode,"XXX");
		}       
	}
	if(0 != strcmp(vars.ntscode,"XXX")) {
		strcat(result,vars.ntscode);
		strcat(result, "    0");     /* add tag code */
		for(i = 0; i<4; i++) {      /* add x, y, x, & diam values */
			strcat(result," ");
			if(i != 3)
				sprintf(str,"%7.2f",*coordptr[i]);
			else
				sprintf(str,"%5.2f",*coordptr[i]);
			strcat(result,str);
		}
		strcat(result, "  ");
		if(0 != strncmp(vars.data_type," ", 1)) /* add type if available */
			strcat(result,vars.data_type);
		strcat(result,"\n\0");              /* add newline & terminate */
	}
 }

/******************** majcode1 ***********************/
/* major code 1: draw line to point */

void majcode1(mincode)
int     mincode; 
{
	switch(mincode){    /* select Neurolucida minor code */
		case 1: {
			strcpy(vars.ntscode,"MTO"); /* middle tree origin */
			strcpy(vars.data_type,"dendrite     ");
			break;
		}
		case 2: {                       
			strcpy(vars.ntscode," CP"); /* continuation */
			break;
		}                               
		case 21: {
			strcpy(vars.ntscode,"MTO"); /* axon start */
			strcpy(vars.data_type,"axon start   ");
			break;
		}
		case 22: {
			strcpy(vars.ntscode," CP"); /* axon continuation */
			break;
		}
		case 41: {
			if(0 == somaflag) {               /* soma outline start */
				strcpy(vars.ntscode,"SOS");
				strcpy(vars.data_type,"soma start   ");
			}
			else {                            /* soma outline end */
				strcpy(vars.ntscode,"SOE");
				strcpy(vars.data_type,"soma stop    ");
			}
			somaflag = !somaflag;
			break;
		}
		case 42: {                             
			strcpy(vars.ntscode,"SCP"); /* soma continuation */
			break;          
		}               
		case 61: {
			strcpy(vars.ntscode,"MTO"); /* apical dendrite start */
			strcpy(vars.data_type,"apical dendr.");
			break;
		}
		case 62: {
			strcpy(vars.ntscode," CP"); /* apical dendrite continuation */
			break;
		}
		default: {                      
			strcpy(vars.ntscode,"XXX"); /*  unknown code */
			break;          
		}                       
	}
}
/******************* majcode2 ************************/
/* major code 2: move to point */

void majcode2(mincode)
int mincode;        
{       
	switch(mincode) {    /* select Neurolucida minor code */
		case 1: {
			strcpy(vars.ntscode,"MTO"); /* middle tree origin */
			strcpy(vars.data_type,"dendrite     ");
			break;
		}
		case 2: {                       
			strcpy(vars.ntscode," CP"); /* continuation */
			break;          
		}               
		case 21: {
			strcpy(vars.ntscode,"MTO"); /* axon start */
			strcpy(vars.data_type,"axon start   ");
			break;
		}
		case 22: {
			strcpy(vars.ntscode," CP"); /* axon continuation */
			break;
		}
		case 41: {
			if(0 == somaflag) {               /* soma outline start */
				strcpy(vars.ntscode,"SOS");
				strcpy(vars.data_type,"soma start   ");
			}
			else {                           /* soma outline end */
				strcpy(vars.ntscode,"SOE");
				strcpy(vars.data_type,"soma stop    ");
			}
			somaflag = !somaflag;
			break;
		}
		case 61: {
			strcpy(vars.ntscode,"MTO"); /* apical dendrite start */
			strcpy(vars.data_type,"apical dendr.");
			break;
		}
		case 62: {
			strcpy(vars.ntscode," CP"); /* apical dendrite continuation */
			break;
		}
		default: {                      
			strcpy(vars.ntscode,"XXX"); /*  unknown code */
			break;          
		}                                       
	}
}

/******************** majcode10 ***********************/
/* major code 10: define object */

void majcode10(mincode)
int   mincode; 
{      
	switch (mincode) {    /* select Neurolucida minor code */
		case 1: {
			strcpy(vars.ntscode,"MTO");    /* middle tree origin */
			strcpy(vars.data_type,"dendrite     ");
			break;
		}
		case 5: {  /* branch point */
			strcpy(vars.ntscode," BP");
			break;
		}
		case 21: {
			strcpy(vars.ntscode,"MTO"); /* axon start */
			strcpy(vars.data_type,"axon start   ");
			break;
		}
		case 22: {
			strcpy(vars.ntscode," CP"); /* axon continuation */
			break;
		}
		case 41: {
			if(0 == somaflag) {           /* soma outline start */
				strcpy(vars.ntscode,"SOS");
				strcpy(vars.data_type,"soma start   ");
			}
			else {                        /* soma outline end */
				strcpy(vars.ntscode,"SOE");
				strcpy(vars.data_type,"soma stop    ");
			}
			somaflag = !somaflag;
			break;
		}
		case 61: {
			strcpy(vars.ntscode,"MTO"); /* apical dendrite start */
			strcpy(vars.data_type,"apical dendr.");
			break;
		}
		case 62: {
			strcpy(vars.ntscode," CP"); /* apical dendrite continuation */
			break;
		 }
		case 101:   /* normal end */
		case 104:   /* artifical end high */
		case 105:   /* end incomplete */
		case 108: { /* artifical end low */
			strcpy(vars.ntscode," NE");
			break;
		}
		default: {
			strcpy(vars.ntscode,"XXX");   /*  unknown code */
			break;
		}
	}
	if(0 == strcmp(vars.ntscode," NE"))
		lineincr = 2;
	else
		lineincr = 1;
}

/******************* majcode32 (not used) ************************/
/* major code 32: specification */

void majcode32(mincode)
int     mincode; 
{      
	switch(mincode) {    /* select Neurolucida minor code */
		case 10: {
			strcpy(vars.ntscode,"XXX"); /* depth increment */
			break;
		}
		case 11: {
			strcpy(vars.ntscode,"XXX"); /* parameter */
			break;          
		}                                               
		case 12: {
			strcpy(vars.ntscode,"XXX"); /* lucida parameter */
			break;
		}
		case 13: {
			strcpy(vars.ntscode,"XXX"); /* mid line start */
			break;
		}
		default: {                      
			strcpy(vars.ntscode,"XXX"); /*  unknown code */
			break;          
		}                       
	}
}

/******************* majcode33 (not used) ************************/
/* major code 33: continuing process */

void majcode33(mincode)
int  mincode; 
{      
	switch(mincode) {    /* select Neurolucida minor code */
		case 101: {
			strcpy(vars.ntscode," NE");  /* continuation normal end*/
			break;
		}
		case 104: {
			strcpy(vars.ntscode,"TAE"); /* continuation end high*/
			break;
		}
		case 105: {
			strcpy(vars.ntscode,"MAE"); /* continuation end incomplete*/
			break;
		}
		case 108: {
			strcpy(vars.ntscode,"BAE"); /* continuation end low */
			break;
		}
		default: {
			strcpy(vars.ntscode,"XXX"); /*  unknown code */
			break;
		}
	}

}





