/* @(#)ntsrd.c	1.9 7/24/91 */
/***************************************************************************
  Copyright (C) 1991 Mona Wong, Salk Institute, San Diego, CA  92126
***************************************************************************/
/* This program will translate from the Eutectic file format to the Douglas
   3D file format */

#include "ntsrd.h"

/* Internal functions */
int process_command_line();
void skip_comments();
void tidy_up();
int translate_others();
int translate_soma();

main (argc, argv)
int argc;
char *argv[];
{

  /* Variable declarations and initializations */
  char * type;
  FILE * eutectic_fp, * douglas_fp;

  /* The algorithm is broken up into small modules;
     first, we will check the command line arguments and handle them 
     properly */
  if (!process_command_line (argc, argv, &eutectic_fp, &douglas_fp))
    exit (ERROR);

  /* We will not be storing the comment lines in the Douglas 3D file */
  skip_comments (&eutectic_fp, &douglas_fp);

  /* Our convention is to place the soma information at the top of the
     Douglas 3D file */
  if (!translate_soma (argv, &eutectic_fp, &douglas_fp)) {
    tidy_up (&eutectic_fp, &douglas_fp);
    exit (ERROR);
  }

  /* Next, we will translate the rest of the data lines */
  if (!translate_others (argv, &eutectic_fp, &douglas_fp)) {
    tidy_up (&eutectic_fp, &douglas_fp);
    exit (ERROR);
  }

  tidy_up (&eutectic_fp, &douglas_fp);
  exit (OK);
} /* end: main (argc, argv) */




/*-----------------------------------------------------------------------*/
/* This function will check the # of args on the command line and then
   open the 2 files;
   RETURN : exit status--ERROR if unsuccessful or OK otherwise
*/
int process_command_line (argc, argv, eutectic_fp, douglas_fp)
int argc;
char * argv[];
FILE ** eutectic_fp, ** douglas_fp;
{
  /* Make sure there is the right # of command line arguments */
  if (argc != COMMAND_LINE_ARGS) {
    fprintf (stderr, 
	     "\n  Usage: %s <input eutectic_filename> <output douglas_filename>\n\n",
	     argv[0]);
    return (ERROR);
  }

  /* Now open the files specified */
  fprintf (stdout, "\n  Checking files ... "); 
  fflush (stdout);
  if (!(*eutectic_fp = fopen (argv[1], "r"))) {
    fprintf (stderr, "error (%s)--cannot open file \"%s\"\n\n", 
	     argv[0], argv[1]);
    return (ERROR);
  }
  if (!(*douglas_fp = fopen (argv[2], "w"))) {
    fprintf (stderr, "error (%s)--cannot open file \"%s\"\n\n", 
	     argv[0], argv[2]);
    close (eutectic_fp);
    return (ERROR);
  }
  fprintf (stdout, "done\n");
  return (OK);
} /* end: int process_command_line (argc, argv, eutectic_fp, douglas_fp) */




/*-----------------------------------------------------------------------*/
/* This function simply ignores the Eutectic file's comment; comments are
   considered to be lines before the header.  The header is designated as
   the string 
   " Point  Type  Tag      X       Y       Z  Thick   Name    Attachment"
   RETURN : nothing
*/
void skip_comments (eutectic_fp, douglas_fp)
FILE ** eutectic_fp, ** douglas_fp;
{

  /* Variable declarations and initializations */
  char line[MAX_LINE_LENGTH];

  /* Simply ignore all comment lines before the header */
  fprintf (stdout, "  Processing header ... ");  
  fflush (stdout);
  while ((fgets (line, MAX_LINE_LENGTH, *eutectic_fp) != NULL) &&
	 strncmp (eutectic_header, line, strlen (eutectic_header)))
    ;
  fprintf (*douglas_fp, "%s\n", douglas_header);
  fprintf (stdout, "done\n");
} /* end: void skip_comments (eutectic_fp, douglas_fp) */




/*-----------------------------------------------------------------------*/
/* This function will search the eutectic file for the soma informatin and
   translate this information first into the douglas file.  Soma data lines
   start with a line of point type SOS and end with a line of point type
   SOE.  We are doing this because we are using the convention that the 
   soma information will always come at the beginning of the file (this 
   is so that the associated frame file for tndp display will be correct)
   RETURN : exit status--ERROR if unsuccessful or OK otherwise
*/
int translate_soma (argv, eutectic_fp, douglas_fp)
char * argv[];
FILE ** eutectic_fp, ** douglas_fp;
{

  /* Variable declaration and initializations */
  char line[MAX_LINE_LENGTH], type[5];
  float x, y, z, diam;
  int cont = YES, found = NO;
  long eutectic_fp_offset;

  /* We will save the current eutectic_fp location so we can return to it
     for the second pass through the data */
  eutectic_fp_offset = ftell (*eutectic_fp);

  fprintf (stdout, "  Translating ... ");  
  fflush (stdout);

  while (fgets (line, MAX_LINE_LENGTH, *eutectic_fp) != NULL) {

    /* If the line isn't blank */
    if (line[0] != '\n') {
      sscanf (line, "%*d %s %*d %f %f %f %f", type, &x, &y, &z, &diam);

      /* If it is the soma, set the found flag */
      if (!strcmp (type, "SOS")) {
	strcpy (type, "mto");
	found = YES;
      }

      /* Else, if it is the end of the soma, unset the cont(inue) flag */
      else if (!strcmp (type, "SOE")) {
	strcpy (type, "ne");
	cont = NO;
      }

      /* Else, if the point type is "scp" */
      else if (!strcmp (type, "SCP"))
	strcpy (type, "scp");


      /* Else, if found is set and we found a point type that is not
	 "SOS", "SOE", or "SCP", then there is an error */
      else if (found) {
	fprintf (stderr, 
		 "error (%s)--reading soma; translation terminated!\n\n", 
		 argv[0]);
	return (ERROR);
      }

      /* If found is set, then we will output the translated line; note 
	 that the Eutectic points are in microns and the Douglas format 
	 is one order magnitude bigger */
      if (found)
	fprintf (*douglas_fp, "%s\t%g\t%g\t%g\t%g\n", type, 
		 x * DOUGLAS_FACTOR, y * DOUGLAS_FACTOR, z * DOUGLAS_FACTOR, 
		 diam * DOUGLAS_FACTOR);

      /* If cont(inue) is unset, then we will leave */
      if (!cont)
	break;
    } /* end: if (line[0] != '\n') */
  } /* end: while (fgets (line, MAX_LINE_LENGTH, *eutectic_fp) != NULL) */

  /* If the found flag is not set, then we didn't find the beginning of
     the soma information ("SOS") and this is an error condition */
  if (!found) {
    fprintf (stderr, 
	     "error (%s)--didn't find soma begin point type \"SOS\";\n",
	     argv[0]);
    fprintf (stderr, "\t\t  translation terminated!\n\n");
    return (ERROR);
  }

  /* Now, restore eutectic_fp's previously saved offset */
  if (!fseek (*eutectic_fp, eutectic_fp_offset, BEGINNING))
    return (OK);

  /* Else, there is error in attempting to rewind the file */
  else {
    fprintf (stderr, 
	     "error (%s)--rewinding \"%s\"; translation terminated!\n\n", 
		 argv[0], argv[1]);
    return (ERROR);
  }
} /* end: int translate_soma (argv, eutectic_fp, douglas_fp) */




/*-----------------------------------------------------------------------*/
/* This function will translate the non-soma data lines from the eutectic
   file to the douglas file accordingly to some simple rules.  Some lines
   will be ignored since they are not really needed
   RETURN : exit status--ERROR if unsuccessful or OK otherwise
*/
int translate_others (argv, eutectic_fp, douglas_fp)
char * argv[];
FILE ** eutectic_fp, ** douglas_fp;
{

  /* Variable declarations */
  char line[MAX_LINE_LENGTH], type[5];
  float x, y, z, diam;
  int i, warning = NO;

  while (fgets (line, MAX_LINE_LENGTH, *eutectic_fp) != NULL) {

    /* If the line isn't blank */
    if (line[0] != '\n') {
      sscanf (line, "%*d %s %*d %f %f %f %f", type, &x, &y, &z, &diam);

      /* Now translate the point types; see the documentation for more 
	 explanation */

      /* First, ignore all soma stuff (all data lines of point type
	 "SOS", "SOE", and "SCP") */
      if (strcmp (type, "SOS") && strcmp (type, "SOE") &&
	  strcmp (type, "SCP")) {

	/* If not soma stuff, then simply translated to its own lower case */
	if (!strcmp (type, "MTO") || !strcmp (type, "BP") || 
	    !strcmp (type, "NE") || !strcmp (type, "MAE") ||
	    !strcmp (type, "TAE") || !strcmp (type, "BAE"))
	  for (i = 0; i < strlen (type); i++)
	    type[i] = tolower (type[i]);

	/* These are translated to their equivalent lower case */
	else if (!strcmp (type, "CP") || !strcmp (type, "SB") 
		 || !strcmp (type, "FS"))
	  strcpy (type, "dcp");
	else if (!strcmp (type, "ES"))
	  strcpy (type, "ne");
	else {
	  warning = YES;
	  fprintf (stderr, 
		   "\n\twarning (%s)--unrecognized point type \"%s\" ignored", 
		   argv[0], type);
	}

	/* Now, output the translated line; note that the Eutectic points
	   are in microns and the Douglas format is one magnitude order
	   bigger */
	fprintf (*douglas_fp, "%s\t%g\t%g\t%g\t%g\n", type, 
		 x * DOUGLAS_FACTOR, y * DOUGLAS_FACTOR, z * DOUGLAS_FACTOR, 
		 diam * DOUGLAS_FACTOR);
      } /* end: if (strcmp (type, "SOS") && strcmp (type, "SOE") ... */
    } /* end: if (line[0] != '\n') */
  } /* end: while (fgets (line, MAX_LINE_LENGTH, *eutectic_fp) != NULL) */

  /* Some formatting considerations */
  if (warning)
    fprintf (stdout, "\n\tdone\n");
  else
    fprintf (stdout, "done\n");
  return (NULL);
} /* end: char * translate_others (argv, eutectic_fp, douglas_fp) */




/*-----------------------------------------------------------------------*/
void tidy_up (eutectic_fp, douglas_fp)
FILE ** eutectic_fp, ** douglas_fp;
{
  close (*eutectic_fp);
  close (*douglas_fp);
  fprintf (stdout, "\n");
} /* end: void tidy_up (eutectic_fp, douglas_fp) */

